"""
This module defines the basic object of the document object model
(DOM).

"""
import os
import sys
from cStringIO import StringIO
LC = sys.modules['lexor.core']
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-instance-attributes


def _write_node_info(node, strf):
    """`Node` helper function to write the node information in
    __repr__. """
    strf.write('%s%s' % ('    '*node.level, node.name))
    if not isinstance(node, LC.Element):
        strf.write('[0x%x]' % id(node))
    else:
        att = ' '.join(['%s="%s"' % (k, v) for k, v in node.items()])
        strf.write('[0x%x' % id(node))
        if att != '':
            strf.write(' %s]' % att)
        else:
            strf.write(']')
    if node.name == '#document':
        strf.write(': (%s:%s:%s)' % (node.uri, node.lang, node.style))
    else:
        strf.write(':')
    direction = 'r'
    if isinstance(node, LC.CharacterData):
        strf.write(' %r' % node.data)
    elif node.child:
        direction = 'd'
    strf.write('\n')
    return direction


def _set_owner_and_level(node, owner, level):
    """Helper method for increase_child_level. """
    if node.owner is not owner:
        if isinstance(node, LC.Element) and 'id' in node:
            try:
                del node.owner.id_dict[node['id']]
            except (KeyError, AttributeError):
                pass
            if owner:
                owner.id_dict[node['id']] = node
        node.owner = owner
    node.level = level
    if node.name == '#document':
        node.level = level - 1


class Node(object):
    """Primary datatype for the entire Document Object Model. """
    __slots__ = ('name', 'owner', 'parent', 'index',
                 'prev', 'next', 'child', 'level')

    def __init__(self):
        """Initializes all data descriptors to ``None``. Each
        descriptor has an associated `READ-ONLY` property. Read the
        comment on each property to see what each descriptor
        represents. """
        self.name = None
        self.owner = None
        self.parent = None
        self.index = None
        self.prev = None
        self.next = None
        self.child = None
        self.level = 0

    @property
    def node_name(self):
        """
        .. admonition:: Read-Only Property
            :class: note

            The name of this node. Its value depends on the node
            type. This property is associated with the attribute
            ``name``. """
        return self.name

    @property
    def owner_document(self):
        """
        .. admonition:: Read-Only Property
            :class: note

            The :class:`~lexor.core.elements.Document` in which this
            node resides. This property is associated with the
            attribute ``owner``. """
        return self.owner

    @property
    def parent_node(self):
        """
        .. admonition:: Read-Only Property
            :class: note

            The parent of this node. If the node has been just
            created or removed from a ``Document`` then this property
            is ``None``. This property is associated with the
            attribute ``parent``. """
        return self.parent

    @property
    def node_index(self):
        """
        .. admonition:: Read-Only Property
            :class: note

            The number of preceding siblings.

            >>> x is x.parent_node[x.node_index]
            True

            This property is associated with the attribute ``index``.
            """
        return self.index

    @property
    def node_level(self):
        """
        .. admonition:: Read-Only Property
            :class: note

            The nodes level of containtment in a
            :class:`~lexor.core.elements.Document` object.

            This property is associated with the attribute ``level``.
            """
        return self.level

    @property
    def element_index(self):
        """
        .. admonition:: Read-Only Property
            :class: note

            The number of preceding element siblings."""
        index = 0
        crt = self
        while crt.prev is not None:
            crt = crt.prev
            if isinstance(crt, LC.Element):
                index += 1
        return index

    @property
    def previous_sibling(self):
        """
        .. admonition:: Read-Only Property
            :class: note

            The node immediately preceding this node. If this
            property is not `None` then

            >>> x.previous_sibling <==> x.parent_node[x.node_index - 1]

            This property is associated with the attribute
            ``prev``."""
        return self.prev

    @property
    def next_sibling(self):
        """
        .. admonition:: Read-Only Property
            :class: note

            The node immediately following this node. If this
            property is not `None` then

            >>> x.next_sibling <==> x.parent_node[x.node_index + 1]

            This property is associated with the attribute
            ``next``."""
        return self.next

    @property
    def previous_element(self):
        """
        .. admonition:: Read-Only Property
            :class: note

            The last sibling :class:`~lexor.core.elements.Element`
            preceding this node.

        """
        crt = self
        while crt.prev is not None:
            crt = crt.prev
            if isinstance(crt, LC.Element):
                return crt
        return None

    @property
    def next_element(self):
        """
        .. admonition:: Read-Only Property
            :class: note

            The sibling :class:`~lexor.core.elements.Element` after
            this node. """
        crt = self
        while crt.next is not None:
            crt = crt.next
            if isinstance(crt, LC.Element):
                return crt
        return None

    def remove_children(self):
        """Remove all the child nodes. """
        for child in self.child:
            child.disconnect()
        del self.child[:]

    def __repr__(self):
        """>>> x.__repr__() == repr(x)
        True
        """
        strf = StringIO()
        crt = self
        direction = _write_node_info(crt, strf)
        if direction is not 'd':
            return strf.getvalue()
        while True:
            if direction is 'd':
                crt = crt.child[0]
            elif direction is 'r':
                if crt.next is None:
                    direction = 'u'
                    continue
                crt = crt.next
            elif direction is 'u':
                if crt.parent is self:
                    break
                if crt.parent.next is None:
                    crt = crt.parent
                    continue
                crt = crt.parent.next
            direction = _write_node_info(crt, strf)
        return strf.getvalue()

    def __str__(self):
        """>>> x.__str__() == str(x)
        True
        """
        if self.owner is None:
            style = 'default'
            lang = 'xml'
        else:
            style = self.owner.style
            lang = self.owner.lang
        writer = LC.Writer(lang, style)
        if self.owner is not None and self.owner.defaults is not None:
            for var, val in self.owner.defaults.iteritems():
                writer.defaults[var] = os.path.expandvars(str(val))
        writer.write(self)
        val = str(writer)
        writer.close()
        return val

    def insert_before(self, index, new_child):
        """Inserts `new_child` to the list of children just before
        the child specified by `index`. """
        if not isinstance(new_child, Node):
            new_child = LC.Text(str(new_child))
        elif isinstance(new_child, LC.DocumentFragment):
            msg = "Use extend_before for LC.DocumentFragment Nodes."
            raise TypeError(msg)
        index = self.insert_node_before(index, new_child)
        while index < len(self.child):
            self[index].index = index
            index += 1
        return self

    def extend_before(self, index, new_children):
        """Inserts the contents of an iterable containing nodes just
        before the child specified by `index`. The following are
        equivalent:

        >>> while doc: node.parent.insert_before(index, doc[0])

        >>> node.extend_before(index, doc)

        The second form, however, has a more efficient reindexing
        method. """
        if isinstance(new_children, (list, LC.DocumentFragment)):
            for node in new_children:
                if node.name == '#document' and node.temporary:
                    if self.owner:
                        self.owner.meta.update(node.meta)
                        node.meta = dict()
                    while node:
                        index = self.insert_node_before(index, node[0])
                else:
                    index = self.insert_node_before(index, node)
        else:
            if new_children.name == '#document':
                if new_children.temporary and self.owner:
                    self.owner.meta.update(new_children.meta)
                    new_children.meta = dict()
            while new_children:
                index = self.insert_node_before(index, new_children[0])
        while index < len(self.child):
            self[index].index = index
            index += 1
        return self

    def append_child(self, new_child):
        """Adds the node `new_child` to the end of the list of
        children of this node. If the node is a
        :class:`~lexor.core.elements.DocumentFragment` then it
        appends its child nodes. Returns the calling node. """
        if not isinstance(new_child, Node):
            new_child = LC.Text(str(new_child))
        elif isinstance(new_child, LC.DocumentFragment):
            msg = "Use extend_children for `DocumentFragment` Nodes."
            raise TypeError(msg)
        self.append_child_node(new_child)
        return self

    def extend_children(self, new_children):
        """Extend the list of children by appending children from an
        iterable containing nodes. """
        if isinstance(new_children, (list, LC.DocumentFragment)):
            for node in new_children:
                if node.name == '#document' and node.temporary:
                    if self.owner:
                        self.owner.meta.update(node.meta)
                        node.meta = dict()
                    while node:
                        self.append_child_node(node[0])
                else:
                    self.append_child_node(node)
        else:
            if new_children.name == '#document':
                if new_children.temporary and self.owner:
                    self.owner.meta.update(new_children.meta)
                    new_children.meta = dict()
            while new_children:
                self.append_child_node(new_children[0])
        return self

    def append_after(self, new_child):
        """Place `new_child` after the node. """
        if self.index+1 == len(self.parent):
            self.parent.append_child(new_child)
        else:
            self.parent.insert_before(self.index+1, new_child)

    def append_nodes_after(self, new_children):
        """Place `new_children` after the node. """
        if self.index+1 == len(self.parent):
            self.parent.extend_children(new_children)
        else:
            self.parent.extend_before(self.index+1, new_children)

    def prepend_before(self, new_child):
        """Place `new_child` before the node. """
        self.parent.insert_before(self.index, new_child)

    def prepend_nodes_before(self, new_children):
        """Place `new_children` before the node. """
        self.parent.extend_before(self.index, new_children)

    def normalize(self):
        """Removes empty :class:`~lexor.core.elements.Text` nodes,
        and joins adjacent :class:`~lexor.core.elements.Text`
        nodes."""
        if not self.child:
            return self
        crt = self.child[0]
        while crt is not None:
            if isinstance(crt, LC.Text):
                if crt.data == '':
                    nextnode = crt.next
                    del crt.parent[crt.index]
                    crt = nextnode
                elif isinstance(crt.next, LC.Text):
                    marked_node = crt.next
                    start = marked_node.index
                    while isinstance(marked_node, LC.Text):
                        crt.data += marked_node.data
                        end = marked_node.index
                        marked_node = marked_node.next
                    crt = marked_node
                    del self[start:end+1]
                else:
                    crt = crt.next
            else:
                crt = crt.next
        return self

    def __len__(self):
        """Return the number of child nodes.

        >>> x.__len__() == len(x)
        True

        """
        if self.child is None:
            return 0
        return len(self.child)

    def __getitem__(self, i):
        """Return the `i`-th child of this node.

        >>> x.__getitem__(i) <==> x[i]
        >>> x.__getitem__(slice(i, j)) <==> x[i:j]
        >>> x.__getitem__(slice(i, j, dt)) <==> x[i:j:dt]

        When using a slice, the ``__getitem__`` function will return
        a list with references to the requested nodes. """
        return self.child[i]

    def _get_indices(self, i):
        """PRIVATE-METHOD: Returns a slice and the range of indices
        to be replaced. `i` is assumed to be a slice or an int. """
        if isinstance(i, int):
            if i < 0:
                i += len(self.child)
                i = slice(i, i + 1)
            else:
                i = slice(i, i + 1)
        return i, xrange(*i.indices(len(self.child)))

    def __delitem__(self, index):
        """Delete child nodes.

        >>> x.__delitem__(index) <==> del x[index]
        >>> x.__delitem__(slice(i, j)) <==> del x[i:j]
        >>> x.__delitem__(slice(i, j, dt)) <==> del x[i:j:dt]

        """
        index, indices = self._get_indices(index)
        if index.step is None or index.step > 0:
            indices = reversed(indices)
        for index in indices:
            self.child[index].disconnect()
            del self.child[index]
            if index > 0:
                try:
                    self.child[index].set_prev(self.child[index - 1])
                except IndexError:
                    self.child[index - 1].next = None
            elif self.child:
                self.child[index].prev = None
        for i in xrange(index, len(self.child)):
            self.child[i].index = i

    def __setitem__(self, index, node):
        """Replace child nodes.

        >>> x.__setitem__(index) = node <==> x[index] = node
        >>> x.__setitem__(slice(i, j)) = dfrag <==> x[i:j] = dfrag
        >>> x.__setitem__(slice(i, j, dt)) = dfrag <==> x[i:j:dt] = dfrag

        When using slices the nodes to be assigned to the indices
        need to be contained in a
        :class:`~lexor.core.elements.DocumentFragment` node. This
        function does not support insertion as the regular slice for
        list does. To insert use a node use :meth:`insert_before` or
        :meth:`append_after`."""
        indices = self._get_indices(index)
        if not isinstance(node, Node):
            raise TypeError("items must be Nodes")
        if node is self:
            raise TypeError("A node cannot have itself as a child.")
        if not isinstance(node, LC.DocumentFragment):
            # Better take a look at this, DocFrag needs more development
            nodes = LC.DocumentFragment(node)
        else:
            nodes = node
        if len(indices) != len(nodes):
            msg = "attempt to assign sequence of size %d to extended" \
                  "slice of size %d" % (len(nodes), len(indices))
            raise ValueError(msg)
        for i in xrange(len(nodes)):
            index = indices[i]
            node = nodes[i]
            if node.parent is self:
                raise ValueError("Node is already the child at index %d" %
                                 node.index)
            if node.parent is not None:
                del node.parent[node.index]
            # Disconnect the current Node
            self.child[index].disconnect()
            # Assign and connect the new Node
            self.child[index] = node
            node.set_parent(self, index)
            try:
                node.set_next(self.child[index + 1])
            except IndexError:
                pass
            try:
                node.set_prev(self.child[index - 1])
            except IndexError:
                pass
        return node

    def get_nodes_by_name(self, name):
        """Return a ``list`` of child nodes that have the given
        `name`. """
        nodes = []
        if not self.child:
            return nodes
        crt = self
        direction = 'd'
        while True:
            if direction is 'd':
                crt = crt.child[0]
            elif direction is 'r':
                if crt.next is None:
                    direction = 'u'
                    continue
                crt = crt.next
            elif direction is 'u':
                if crt.parent is self:
                    break
                if crt.parent.next is None:
                    crt = crt.parent
                    continue
                crt = crt.parent.next
            if crt.name == name:
                nodes.append(crt)
            if crt.child:
                direction = 'd'
            else:
                direction = 'r'
        return nodes

    def iter_child_elements(self):
        """Iterate over child nodes which are
        :class:`~lexor.core.elements.Element` instances.

        >>> for ele in node.iter_child_elements(): ...

        """
        for k in self.child:
            if isinstance(k, LC.Element):
                yield k

    def set_parent(self, parent, index):
        """
        .. admonition:: Helper Method
            :class: caution

            Modifies the parent node and takes care of the child node
            levels. """
        self.parent = parent
        self.index = index
        if self.name in ['#document', '#document-fragment']:
            self.level = parent.level
        else:
            self.level = parent.level + 1
        self.owner = parent.owner
        if self.owner and isinstance(self, LC.Element) and 'id' in self:
            self.owner.id_dict[self['id']] = self
        self.increase_child_level()

    def disconnect(self):
        """
        .. admonition:: Helper Method
            :class: caution

            Reset its attributes.
        """
        if self.owner and isinstance(self, LC.Element):
            try:
                del self.owner.id_dict[self['id']]
            except (KeyError, AttributeError):
                pass
        self.owner = None
        self.parent = None
        self.index = None
        self.prev = None
        self.next = None
        if self.name in ['#document', '#document-fragment']:
            self.level = -1
        else:
            self.level = 0
        self.increase_child_level()

    def set_prev(self, node):
        """
        .. admonition:: Helper Method
            :class: caution

            Sets the ``prev`` attribute.
        """
        self.prev = node
        node.next = self

    def set_next(self, node):
        """
        .. admonition:: Helper Method
            :class: caution

            Sets the ``next`` attribute.
        """
        self.next = node
        node.prev = self

    def increase_child_level(self):
        """
        .. admonition:: Helper Method
            :class: caution

            Sets the level of the child nodes.
        """
        if self.child:
            crt = self
            direction = 'd'
            level = self.level
            owner = self.owner
        else:
            return
        while True:
            if direction is 'd':
                level += 1
                crt = crt.child[0]
            elif direction is 'r':
                if crt.next is None:
                    direction = 'u'
                    continue
                crt = crt.next
            elif direction is 'u':
                level -= 1
                if crt.parent is self:
                    break
                if crt.parent.next is None:
                    crt = crt.parent
                    continue
                crt = crt.parent.next
            _set_owner_and_level(crt, owner, level)
            if crt.child:
                direction = 'd'
            else:
                direction = 'r'

    def append_child_node(self, new_child):
        """
        .. admonition:: Helper Method
            :class: caution

            Use this method to insert a node at a the end of the
            child list. See :meth:`append_child` and
            :meth:`extend_children` to see this method in
            action.
        """
        if new_child.parent is not None:
            del new_child.parent[new_child.index]
        self.child.append(new_child)
        new_child.set_parent(self, len(self.child) - 1)
        try:
            new_child.set_prev(self.child[-2])
        except IndexError:
            pass

    def insert_node_before(self, index, new_child):
        """
        .. admonition:: Helper Method
            :class: caution

            Insert a `new_child` at a given `index`. See
            :meth:`insert_before` and :meth:`extend_before` to see
            this method in action."""
        if new_child.parent is not None:
            del new_child.parent[new_child.index]
        self.child.insert(index, new_child)
        new_child.set_parent(self, index)
        if index > 0:
            new_child.set_prev(self.child[index-1])
        try:
            new_child.set_next(self.child[index+1])
        except IndexError:
            pass
        self[index].index = index
        index += 1
        return index



"""logging

The logging module provides a basic object to store information about
an event and a ``Logger`` which creates these events. The module
provides the global variable ``L`` which may be used to record events
during the usage of lexor. For more information on ``L`` see the
documentation on :class:`~.Logger`.

"""

import sys
import traceback
from datetime import datetime
from inspect import currentframe, getframeinfo


class LogMessage(object):
    """Simple object to store an event information."""

    def __init__(self, fname, func_name, lineno, kind, msg, lvl, exception=None):
        self.date = datetime.now()
        self.file_name = fname
        self.func_name = func_name
        self.line_number = lineno
        self.kind = kind
        self.message = msg
        self.exception = exception
        self.exceptionTraceback = None
        self.level = lvl
        if exception is not None:
            self.exceptionTraceback = traceback.format_exc()


    def __repr__(self):
        msg = '[%s][%s][%s:%d] => %s' % (
            self.kind, self.func_name, self.file_name,
            self.line_number, self.message
        )
        if self.exception is not None:
            msg += '\n' + self.exceptionTraceback
        return msg


class Logger(object):
    """Provides methods function to tag an event. """

    def __init__(self):
        self.on = False
        self.history = []

    def enable(self):
        self.on = True

    def disable(self):
        self.on = False

    def _push(self, cfr, kind, msg, *args, **kwargs):
        f_back = cfr.f_back
        fname = getframeinfo(f_back).filename
        parts = fname.split('lexor/lexor/')
        if len(parts) == 2:
            fname = parts[1]
        lineno = f_back.f_lineno
        func_name = f_back.f_code.co_name
        exception = kwargs.get('exception', None)
        level = kwargs.get('level', '=')
        self.history.append(LogMessage(
            fname, func_name, lineno,
            kind, msg % args, level, exception
        ))

    def log(self, msg, *args, **kwargs):
        if self.on:
            self._push(currentframe(), 'LOG', msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        if self.on:
            self._push(currentframe(), 'INFO', msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        if self.on:
            self._push(currentframe(), 'DEBUG', msg, *args, **kwargs)

    def warn(self, msg, *args, **kwargs):
        self._push(currentframe(), 'WARN', msg, *args, **kwargs)
        sys.stderr.write('%r\n' % self.history[-1])

    def error(self, msg, *args, **kwargs):
        self._push(currentframe(), 'ERROR', msg, *args, **kwargs)
        if not self.on:
            self.history[-1].exception = None
        sys.stderr.write('%r\n' % self.history[-1])

    def msg(self, kind, msg, *args, **kwargs):
        if self.on:
            self._push(currentframe(), kind, msg, *args, **kwargs)

    def __repr__(self):
        return '\n'.join([repr(x) for x in self.history])


L = Logger()



"""
Here we can find functions to communicate with the cloud to find
lexor packages in Github, register them and delete them.

"""
import os
import six
import json
import textwrap
import httplib
from imp import load_source
from lexor.command import disp
from lexor.command import config, LexorError

DESC = {'main': """
communicate with the cloud. This will allow you to register a style
that is being hosted in Github. You must provide a Github access
token in order to use this command. To create a Github token see:

https://help.github.com/articles/creating-an-access-token-for-command-line-use/

If you wish to avoid typing the Github token you may save the token
to the environmental variable `$GITHUB_ACCESS_TOKEN`.

""", 'register': """
register the style specified by the input file. Make sure that the
style is an existing repository in Github and to provide your Github
access code to the cloud.

""", 'delete': """
delete the style specified by the input file. Make sure that the
style is an existing repository in Github and to provide your Github
access code to the cloud.

""", 'list': """
list available styles. This has to be a string with the following
format:

    lang.type.[to_lang].style

where ``to_lang`` is only required if ``type`` is set to converter.
If you want to display all the available options then you may use
``_``. For instance to search all the available writers

    _.writer

"""}


def add_parser(subp, fclass):
    """Add a parser to the main subparser. """
    main = subp.add_parser('cloud', help='communicate with the cloud',
                           formatter_class=fclass,
                           description=textwrap.dedent(DESC['main']))
    main.add_argument('--access-token', type=str, dest='token',
                      metavar='TOKEN',
                      help='github access token')

    sub = main.add_subparsers(title='subcommands',
                              dest='subparser_name',
                              help='additional help',
                              metavar='<command>')

    sub.add_parser('register', help='register a style',
                   formatter_class=fclass,
                   description=textwrap.dedent(DESC['register']))

    sub.add_parser('delete', help='register a style',
                   formatter_class=fclass,
                   description=textwrap.dedent(DESC['delete']))

    tmp = sub.add_parser('list', help='list available styles',
                         formatter_class=fclass,
                         description=textwrap.dedent(DESC['list']))
    tmp.add_argument('style', type=str,
                     help="style to search: lang.type.[to_lang].style")


def run():
    """Perform a cloud operation. """
    arg = config.CONFIG['arg']
    if arg.subparser_name == 'list':
        _list(arg)
        exit(0)
    path = os.path.abspath(arg.inputfile)
    if '.py' not in path:
        path = '%s.py' % path
    try:
        mod = load_source("tmp-mod", path)
    except IOError:
        raise LexorError('not a valid module')
    if not hasattr(mod, 'INFO'):
        raise LexorError("module does not have `INFO`")
    token = ''
    if arg.token:
        token = arg.token
    elif mod.INFO['git']['host'] == 'github':
        token = os.environ.get('GITHUB_ACCESS_TOKEN', '')
    if token == '':
        msg = 'github token is required to %s the style in the cloud'
        raise LexorError(msg % arg.subparser_name)
    func = {
        'register': _register,
        'delete': _delete,
    }
    func[arg.subparser_name](arg, mod.INFO, token)


def _register(_, info, token):
    """Register a style in the cloud. """
    inputs = {
        'style': info['style'],
        'to_lang': str(info['to_lang']),
        'type': info['type'],
        'lang': info['lang'],
        'repo': info['git']['repo'],
        'user': info['git']['user'],
        'host': info['git']['host'],
        'token': token
    }
    ans = cloud_request('register', inputs)
    if isinstance(ans, six.string_types):
        raise LexorError(ans)
    else:
        disp("registration successful:\n")
        for key in ans:
            disp('  %s: %s\n' % (key, ans[key]))


def _delete(_, info, token):
    """Register a style in the cloud. """
    inputs = {
        'repo': info['git']['repo'],
        'user': info['git']['user'],
        'host': info['git']['host'],
        'token': token
    }
    ans = cloud_request('delete', inputs)
    if isinstance(ans, six.string_types):
        raise LexorError(ans)
    else:
        disp("deletion successful:\n")
        for key in ans:
            disp('  %s: %s\n' % (key, ans[key]))


def _list(arg):
    """List the available styles. """
    inputs = {}
    styles = arg.style.split('.')
    size = len(styles)
    if size > 0 and styles[0] != '_':
        inputs['lang'] = styles[0]
    if size > 1 and styles[1] != '_':
        if styles[1] not in ['parser', 'writer', 'converter']:
            raise LexorError("'%s' is not a valid type" % styles[1])
        inputs['type'] = styles[1]
    if size > 2 and styles[2] != '_':
        inputs['to_lang' if styles[1] == 'converter' else 'style'] = styles[2]
    if size > 3 and styles[3] != '_':
        inputs['style'] = styles[3]
    ans = cloud_request('match', inputs)
    if isinstance(ans, six.string_types):
        raise LexorError(ans)
    if not ans:
        disp('no matches found\n')
    for item in ans:
        name = '%s.%s' % (item['lang'], item['type'])
        if item['to_lang'] != 'None':
            name += '.%s' % item['to_lang']
        name += '.%s' % item['style']
        url = "https://github.com/%s/%s" % (item['user'], item['repo'])
        print '{0:35} {1:>35}'.format(name, url)


def cloud_request(request, data):
    """Make a request to the cloud. """
    data = json.dumps(data)
    lexor_key = "aUQhbr8UXCHooj41tY9YtbGyo9L7Uv3yN7sEAPDh"
    rest_api_key = "RQZuvdQexH7HtnJWVWTlREJbWrodiTb2uitSqTLL"
    connection = httplib.HTTPSConnection('api.parse.com', 443)
    connection.connect()
    connection.request('POST', '/1/functions/'+request, data, {
        "X-Parse-Application-Id": lexor_key,
        "X-Parse-REST-API-Key": rest_api_key,
        "Content-Type": "application/json"
    })
    result = json.loads(connection.getresponse().read())
    if 'error' in result:
        return result['error']
    else:
        return result['result']



"""
This module provides functions to load the different parsers, writers
and converters. It defines the list ``LEXOR_PATH`` which is an array
of paths where lexor looks for the parsing, writing and converting
styles. You may specify more paths by either directly appending to
this path while using lexor as a python module or by editing the
enviromental variable ``LEXORPATH`` and appending paths to it
separating them by a colon ``:``.

"""

import os
import sys
import site
import textwrap
from os.path import splitext, abspath
from imp import load_source
from glob import iglob, glob
from lexor.util.logging import L
from lexor.command import config, disp, LexorError


DEFAULTS = {
    '_': 'default',
    'md': 'markdown',
    'mdown': 'markdown',
    'mkdn': 'markdown',
    'mkd': 'markdown',
    'mdwn': 'markdown',
    'mdtxt': 'markdown',
    'mdtext': 'markdown',
    'text': 'markdown',
    'lex': 'lexor',
    'pyhtml': 'html',
    'pyxml': 'xml',
}
DESC = """
Show all the lexor modules which are installed in the directories
specified in ``LEXORPATH``.

"""
try:
    LEXOR_PATH = [
        './lexor_modules',
        '%s/lib/lexor_modules' % site.getuserbase(),
        '%s/lib/lexor_modules' % sys.prefix
    ]
except AttributeError:
    LEXOR_PATH = [
        './lexor_modules',
        '%s/lib/lexor_modules' % sys.prefix
    ]

if 'LEXORPATH' in os.environ:
    LEXOR_PATH = os.environ['LEXORPATH'].split(':') + LEXOR_PATH


def add_parser(subp, fclass):
    """
    .. admonition:: Command Line Utility Function
        :class: warning

        Add a parser to the main subparser.
    """
    subp.add_parser('lang', help='see available styles',
                    formatter_class=fclass,
                    description=textwrap.dedent(DESC))


def _handle_kind(paths):
    """Helper function for _handle_lang. """
    styles = dict()
    for path in paths:
        L.info('... processing %s', path)
        pattern = '%s/*.py' % path
        tmp = [ele for ele in glob(pattern)]
        for module_path in tmp:
            try:
                mod_name = '__tmp%d__' % _handle_kind.loaded
                module = load_source(mod_name, module_path)
                _handle_kind.loaded += 1
            except ImportError:
                continue
            style = os.path.basename(module_path)[:-3]
            L.info('      - %s: %r', style, module)
            if style not in styles:
                styles[style] = []
            styles[style].append(module)
    for style in sorted(styles.keys()):
        disp('        [*] %s ->\n' % style)
        for module in styles[style]:
            info = module.INFO
            msg = '               %s: %s\n'
            disp(msg % (info['ver'], info['path']))
_handle_kind.loaded = 0


def _handle_lang(path):
    """Helper function for run. """
    for kind in sorted(path.keys()):
        disp('    %s:\n' % kind)
        _handle_kind(path[kind])


def run():
    """
    .. admonition:: Command Line Utility Function
        :class: warning

        Run the command.
    """
    L.info('searching for lexor modules ...')
    paths = []
    for base in LEXOR_PATH:
        paths += glob('%s/*' % base)
    path = dict()
    for loc in paths:
        L.info('searching in %r', loc)
        kind = os.path.basename(loc)
        try:
            name, kind = kind.split('.', 1)
        except ValueError:
            continue
        if name not in path:
            path[name] = dict()
        if kind not in path[name]:
            path[name][kind] = [loc]
        else:
            path[name][kind].append(loc)
    for lang in sorted(path.keys()):
        disp('%s:\n' % lang)
        _handle_lang(path[lang])
        disp('\n')


def _get_info(cfg, type_, lang, style, to_lang=None):
    """Helper function for get_style_module. """
    if style == '_':
        style = 'default'
    if lang in cfg['lang']:
        lang = cfg['lang'][lang]
    if to_lang:
        if to_lang in cfg['lang']:
            to_lang = cfg['lang'][to_lang]
        key = '%s.%s.%s.%s' % (lang, type_, to_lang, style)
        name = '%s.%s.%s/%s' % (lang, type_, to_lang, style)
        modname = 'lexor-lang_%s_%s_%s_%s'
        modname %= (lang, type_, to_lang, style)
    else:
        key = '%s.%s.%s' % (lang, type_, style)
        name = '%s.%s/%s' % (lang, type_, style)
        modname = 'lexor-lang_%s_%s_%s' % (lang, type_, style)
    return key, name, modname


def get_style_module(type_, lang, style, to_lang=None):
    """Return a parsing/writing/converting module. """
    cfg = config.get_cfg(['lang', 'develop'])
    config.update_single(cfg, 'lang', DEFAULTS)
    key, name, modname = _get_info(cfg, type_, lang, style, to_lang)
    L.info('searching for %s', name)
    if 'develop' in cfg:
        try:
            path = cfg['develop'][key]
            if path[0] != '/':
                path = '%s/%s' % (config.CONFIG['path'], path)
            try:
                module = load_source(modname, path)
            except IOError:
                msg = 'Unable to load module in development: %s'
                raise LexorError(msg % path)
            ver = module.INFO['ver']
            L.info('... developing v%s from %s', ver, path)
            return module
        except KeyError:
            pass
    for base in LEXOR_PATH:
        path = '%s/%s.py' % (base, name)
        try:
            module = load_source(modname, path)
            ver = module.INFO['ver']
            L.info('... found v%s in %r', ver, base)
            return module
        except IOError:
            L.info('... searched in %r', base)
    raise ImportError('lexor module not found: %s' % name)


def load_mod(modbase, dirpath):
    """Return a dictionary containing the modules located in
    `dirpath`. The name `modbase` must be provided so that each
    module may have a unique identifying name. The result will be a
    dictionary of modules. Each of the modules will have the name
    ``"modbase_modname"`` where modname is a module in the
    directory."""
    mod = dict()
    for path in iglob('%s/*.py' % dirpath):
        if 'test' not in path:
            module = path.split('/')[-1][:-3]
            modname = '%s_%s' % (modbase, module)
            mod[module] = load_source(modname, path)
    return mod


def load_aux(info):
    """Wrapper around :func:`load_mod` for easy use when developing
    styles. The only parameter is the dictionary ``INFO`` that needs
    to be defined with every style. ``INFO`` is returned by the
    :func:`lexor.init` function."""
    dirpath = splitext(abspath(info['path']))[0]
    if info['to_lang']:
        modbase = 'lexor-lang_%s_converter_%s' % (info['lang'],
                                                  info['to_lang'])
    else:
        modbase = 'lexor-lang_%s_%s_%s' % (info['lang'],
                                           info['type'],
                                           info['style'])
    return load_mod(modbase, dirpath)


def load_rel(path, module):
    """Load relative to a path. If path is the name of a file the
    filename will be dropped."""
    if not os.path.isdir(path):
        path = os.path.dirname(os.path.realpath(path))
    if '.py' in module:
        module = module[1:-3]
    fname = '%s/%s.py' % (path, module)
    return load_source('load-rel-%s' % module, fname)


def map_explanations(mod, exp):
    """Helper function to create a map of msg codes to explanations
    in the lexor language modules. """
    if not mod:
        return
    for mod_name, module in mod.iteritems():
        exp[mod_name] = dict()
        codes = module.MSG.keys()
        for index in xrange(len(module.MSG_EXPLANATION)):
            sub = len(codes) - 1
            while sub > -1:
                code = codes[sub]
                if code in module.MSG_EXPLANATION[index]:
                    del codes[sub]
                    exp[mod_name][code] = index
                sub -= 1
            if not codes:
                break



"""lexor setup script"""

import imp
import os.path as pt
from setuptools import setup


def get_version():
    """Get version & version_info without importing lexor.__init__ """
    path = pt.join(pt.dirname(__file__), 'lexor', '__version__.py')
    mod = imp.load_source('lexor_version', path)
    return mod.VERSION, mod.VERSION_INFO

VERSION, VERSION_INFO = get_version()

DESCRIPTION = "Document converter implemented in python."
LONG_DESCRIPTION = open(pt.join(pt.dirname(__file__), 'README.rst')).read()
LONG_DESCRIPTION += open(pt.join(pt.dirname(__file__), 'HISTORY.rst')).read()

DEV_STATUS_MAP = {
    'alpha': '3 - Alpha',
    'beta': '4 - Beta',
    'rc': '4 - Beta',
    'final': '5 - Production/Stable'
}
if VERSION_INFO[3] == 'alpha' and VERSION_INFO[4] == 0:
    DEVSTATUS = '2 - Pre-Alpha'
else:
    DEVSTATUS = DEV_STATUS_MAP[VERSION_INFO[3]]

setup(name='lexor',
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      keywords='lexor markdown html',
      author='Manuel Lopez',
      author_email='jmlopez.rod@gmail.com',
      url='http://lexor.readthedocs.org',
      license='BSD License',
      packages=[
          'lexor',
          'lexor.command',
          'lexor.core',
          ],
      scripts=['bin/lexor'],
      install_requires=[
          'configparser>=3.3.0r2',
          'argcomplete>=0.6.7',
          'nose>=1.3',
          'python-dateutil>=2.2',
          ],
      package_data={
          'lexor.core': ['*.txt'],
          },
      include_package_data=True,
      classifiers=[
          'Development Status :: %s' % DEVSTATUS,
          'License :: OSI Approved :: BSD License',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: POSIX :: Linux',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.7',
          'Topic :: Documentation',
          'Topic :: Communications :: Email',
          'Topic :: Text Processing',
          'Topic :: Text Processing :: Linguistic',
          'Topic :: Text Processing :: Markup',
          ],
     )



"""
This module is trying to simulate jquery selectors. If some code
looks similar to that of the Sizzle CSS Selector engine it is because
the ideas were taken from it.

In short, credit goes to Sizzle_ and CSS for the selector idea.

.. warning::

    This module is far from complete. Some of the functions specified
    here lack functionality.

.. _Sizzle: http://sizzlejs.com/

"""

import re
import sys
import types
from datetime import datetime
from time import mktime
LC = sys.modules['lexor.core']


def get_date():
    """Obtain an integer representation of the date. """
    date = datetime.utcnow()
    return int(mktime(date.timetuple()))


def mark_function(fnc):
    """Mark a function for special use by Sizzle. """
    fnc.expando = True
    return fnc


BOOLEANS = "checked|selected|async|autofocus|autoplay|controls|" + \
           "defer|disabled|hidden|ismap|loop|multiple|open|" + \
           "readonly|required|scoped"
WHITESPACE = "[\\x20\\t\\r\\n\\f]"
CHAR_ENCODING = "(?:\\\\.|[\\w-]|[^\\x00-\\xa0])+"
IDENTIFIER = CHAR_ENCODING.replace("w", "w#")
ATTRIBUTES = "\\[" + WHITESPACE + "*(" + CHAR_ENCODING + ")" + \
             WHITESPACE + "*(?:([*^$|!~]?=)" + WHITESPACE + \
             "*(?:(['\"])((?:\\\\.|[^\\\\])*?)\\3|(" + \
             IDENTIFIER + ")|)|)" + WHITESPACE + "*\\]"
PSEUDOS = ":(" + CHAR_ENCODING + \
          ")(?:\\(((['\"])((?:\\\\.|[^\\\\])*?)\\3|((?:\\\\.|" + \
          "[^\\\\()[\\]]|" + ATTRIBUTES.replace("3", "8") + \
          ")*)|.*)\\)|)"
RCOMMA = re.compile("^" + WHITESPACE + "*," + WHITESPACE + "*")
RCOMBINATORS = re.compile("^" + WHITESPACE + "*([>+~]|" +
                          WHITESPACE + ")" + WHITESPACE + "*")
# Use .findall instead of search or match since this regex had the
# global attribute
RATTRIBUTEQUOTE = re.compile("=" + WHITESPACE + "*([^\\]'\"]*?)" +
                             WHITESPACE + "*\\]")
RPSEUDO = re.compile(PSEUDOS)
RIDENTIFIER = re.compile("^" + IDENTIFIER + "$")
MATCH_EXPR = {
    "ID": re.compile("^#(" + CHAR_ENCODING + ")"),
    "CLASS": re.compile("^\\.(" + CHAR_ENCODING + ")"),
    "TAG": re.compile("^(" + CHAR_ENCODING.replace("w", "w*") + ")"),
    "ATTR": re.compile("^" + ATTRIBUTES),
    "PSEUDO": re.compile("^" + PSEUDOS),
    "CHILD": re.compile("^:(only|first|last|nth|nth-last)-" +
                        "(child|of-type)(?:\\(" + WHITESPACE +
                        "*(even|odd|(([+-]|)(\\d*)n|)" + WHITESPACE +
                        "*(?:([+-]|)" + WHITESPACE + "*(\\d+)|))" +
                        WHITESPACE + "*\\)|)", re.IGNORECASE),
    "bool": re.compile("^(?:" + BOOLEANS + ")$", re.IGNORECASE),
    # For use in libraries implementing .is()
    # We use this for POS matching in `select`
    "needsContext": re.compile("^" + WHITESPACE +
                               "*[>+~]|:(even|odd|eq|gt|lt|nth|" +
                               "first|last)(?:\\(" + WHITESPACE +
                               "*((?:-\\d)?\\d*)" + WHITESPACE +
                               "*\\)|)(?=[^-]|$)", re.IGNORECASE),
}
RQUICKEXPR = re.compile(r'^(?:#([\w-]+)|(\w+)|\.([\w-]+))$')
RUNESCAPE = re.compile("\\\\([\\da-f]{1,6}" + WHITESPACE + "?|(" +
                       WHITESPACE + ")|.)", re.IGNORECASE)
EXPANDO = 'sizzle'+str(get_date())


def _pre_filter_attr(match):
    """function for EXPR['pre_filter']['ATTR']"""
    # match[0] = match[0].replace(runescape, funescape);
    # match[0] = re.sub(runescape, funescape, match[0])
    # Move the given value to match[3] whether quoted or unquoted
    match[2] = match[3] or match[4] or ""
    # match[2] = match[2].replace(runescape, funescape);
    if match[1] == "~=":
        match[2] = " " + match[2] + " "
    return match[:4]


def _pre_filter_child(match):
    """function for EXPR['pre_filter']['CHILD']"""
    match[0] = match[0].lower()
    if match[0].slice[:3] == "nth":
        # nth-* requires argument
        if not match[2]:
            print 'ERROR'
            sys.exit(2)
            # Should raise an error showing match[0]
        # numeric x and y parameters for Expr.filter.CHILD
        # remember that false/true cast respectively to 0/1
        # ML: Possible +=
        if match[4]:
            match[4] = match[5] + (match[6] or 1)
        else:
            match[4] = 2*(match[3] == "even" or match[3] == "odd")
        match[5] = ((match[7] + match[8]) or match[3] == "odd")
    # other types prohibit arguments
    elif match[3]:
        print 'ERROR'
        sys.exit(2)
    return match


def _filter_tag(node_name_selector):
    name = node_name_selector
    if node_name_selector == '*':
        return lambda: True
    else:
        return lambda elem: elem and elem.name.lower() == name


def _filter_class(class_name):
    try:
        pattern = _filter_class.cache[class_name]
        return pattern
    except KeyError:
        pass
    pattern = re.compile("(^|" + WHITESPACE + ")" + class_name + "(" + WHITESPACE + "|$)")
    _filter_class.cache[class_name] = lambda elem: pattern.search(elem['class']) if 'class' in elem else None
    return _filter_class.cache[class_name]
_filter_class.cache = dict()

EXPR = {
    'create_pseudo': mark_function,
    'match': MATCH_EXPR,
    'attr_handle': {},
    'find': {},
    'relative': {
        '>': {'dir': "parent_node", 'first': True},
        ' ': {'dir': "parent_node"},
        '+': {'dir': "previous_sibling", 'first': True},
        '~': {'dir': "previous_sibling"}
    },
    'pre_filter': {
        'ATTR': _pre_filter_attr,
        'CHILD': _pre_filter_child,
    },
    'filter': {
        'TAG': _filter_tag,
        'CLASS': _filter_class,
    }
}


def clone_obj(obj, parser):
    """Utility function to create deep copies of objects used for the
    Selector object. A parser should be given in case the object is a
    string. """
    try:
        return obj.clone_node(True)
    except AttributeError:
        pass
    if hasattr(obj, '__iter__'):
        return [clone_obj(ele, parser) for ele in obj]
    parser.parse(str(obj))
    return parser.doc


def sizzle(selector, context, results=None, seed=None):
    """Function shamelessly borrowed and partially translated to
    python from http://sizzlejs.com/. """
    if results is None:
        results = list()
    if not selector or not isinstance(selector, str):
        return results
    if not isinstance(context, LC.Element):
        return list()
    match = RQUICKEXPR.match(selector)
    if match is not None:  # Shortcuts
        match = match.groups()
        element_id = match[0]
        if element_id:  # sizzle('#ID')
            if context.name == '#document':
                elem = context.get_element_by_id(element_id)
                if elem:
                    results.append(elem)
            elif context.owner:
                elem = context.owner.get_element_by_id(element_id)
                if elem and context.contains(elem):
                    results.append(elem)
        elif match[1]:  # sizzle('TAG')
            results.extend(context.get_nodes_by_name(selector))
        elif isinstance(context, LC.Element):  # sizzle('.CLASS')
            results.extend(context.get_elements_by_class_name(match[2]))
        return results
    return select(selector.strip(), context, results, seed)


def select(selector, context, results, seed):
    """A low-level selection function that works with Sizzle's
    compiled selector functions

    :``selector``: A selector or a pre-compiled selector function
        built with ``Sizzle.compile``
    :``context``: An :class:`lexor.core.elements.Element` object.
    :``results``: A ``list``.
    :``seed``: A ``list`` of elements to match against.
    """
    # compiled = typeof selector === "function" && selector
    # selector = compiled.selector or selector
    match = not seed and tokenize(selector)
    results = results or list()
    if len(match) == 1:
        # Other complicated stuff
        pass
    compile_selector(selector, match)(seed, context, results)
    return results


def matcher_from_tokens(tokens):
    """
    .. warning::

        Missing documentation.
    """
    pass


def matcher_from_group_matchers(element_matchers, set_matchers):
    """
    .. warning::

        Missing documentation.
    """
    pass


def compile_selector(selector, match=None):
    """
    .. warning::

        Missing documentation.
    """
    try:
        return compile_selector.cache[selector]
    except KeyError:
        pass
    set_matchers = list()
    element_matchers = list()
    if match is None:
        match = tokenize(selector)
    i = len(match) - 1
    while i:
        cached = matcher_from_tokens(match[i])
        if cached[EXPANDO]:
            set_matchers.append(cached)
        else:
            element_matchers.append(cached)
    cached = matcher_from_group_matchers(element_matchers, set_matchers)
    compile_selector.cache[selector] = cached
    # cached.selector = selector
    return cached
compile_selector.cache = dict()


def tokenize(selector, parse_only=False):
    """
    .. warning::

        Missing documentation.
    """
    try:
        cached = tokenize.cache[selector]
    except KeyError:
        pass
    else:
        return 0 if parse_only else cached
    so_far = selector
    groups = list()
    pre_filters = EXPR['pre_filter']
    matched = False
    while so_far:
        match = RCOMMA.match(so_far)
        if not matched or match:
            if match:
                match = match.groups()
                so_far = so_far[len(match[0]):] or so_far
            tokens = list()
            groups.append(tokens)
        matched = False
        match = RCOMBINATORS.match(so_far)
        if match:
            matched = match.group(0)
            match = list(match.groups())
            tokens.append({
                'value': matched,
                'type': match[0].strip(),
            })
            so_far = so_far[len(matched):]
        for ftype in EXPR['filter']:
            match = MATCH_EXPR[ftype].match(so_far)
            if match:
                matched = match.group(0)
                match = list(match.groups())
                # match = pre_filters[ftype](match).groups()
                tokens.append({
                    'value': matched,
                    'type': ftype,
                    'matches': match,
                })
                so_far = so_far[len(matched):]
        if not matched:
            break
    if parse_only:
        return len(so_far)
    else:
        tokenize.cache[selector] = groups
        return tokenize.cache[selector]
if not hasattr(tokenize, 'cache'):
    tokenize.cache = dict()


class Selector(object):
    """JQuery like object. """

    def __init__(self, selector, node, results=None):
        self.data = sizzle(selector, node, results)

    def __getitem__(self, k):
        """Return the `k`-th element selected.

        >>> x.__getitem__(k) <==> x[k]

        """
        return self.data[k]

    def __repr__(self):
        """
        >>> x.__repr__() == repr(x)
        True

        """
        result = '\n----------\n'
        for node in self.data:
            result += repr(node)
            result += '\n----------\n'
        return result

    def find(self, selector):
        """Get the descendants of each element in the current set of
        matched elements, filtered by a selector. """
        current = self.data
        self.data = list()
        for node in current:
            sizzle(selector, node, self.data)
        return self

    def contents(self):
        """Get the children of each element in the set of matched
        elements, including text and comment nodes."""
        current = self.data
        self.data = list()
        for node in current:
            if node:
                self.data.extend(node.child)
        return self

    @staticmethod
    def _append(node, content, parser):
        """Helper function to `append` method. """
        if isinstance(content, Selector):
            node.extend_children(content.data)
        elif isinstance(content, LC.Node):
            if (content.name in ['#document', '#document-fragment']
                    and content.temporary):
                node.extend_children(content)
            else:
                node.append_child(content)
        elif hasattr(content, '__iter__'):
            node.extend_children(content)
        else:
            parser.parse(str(content))
            node.extend_children(parser.doc)

    def append(self, *arg, **keywords):
        """Insert content, specified by the parameter, to the end of
        each element in the set of matched elements.

        Should behave similarly as https://api.jquery.com/append/.
        Major difference is in the function. When passing a function
        it should take 2 parameters: node, index. Where node will be
        the current element to which the return value will be
        appended to. """
        info = {
            'lang': 'html',
            'style': 'default',
            'defaults': None,
        }
        for key in keywords:
            info[key] = keywords[key]
        parser = LC.Parser(info['lang'], info['style'], info['defaults'])
        if len(arg) == 1 and isinstance(arg[0], types.FunctionType):
            for num, node in enumerate(self.data):
                self._append(node, arg[0](node, num), parser)
        else:
            for content in arg:
                if isinstance(content, str):
                    parser.parse(content)
                    content = parser.doc
                elif isinstance(content, list):
                    for num in xrange(len(content)):
                        if isinstance(content[num], str):
                            parser.parse(content[num])
                            content[num] = parser.doc
                for i in xrange(len(self.data) - 1):
                    clone = clone_obj(content, parser)
                    self._append(self.data[i], clone, parser)
                if self.data:
                    self._append(self.data[-1], content, parser)

    @staticmethod
    def _prepend(node, content, parser):
        """Helper function to `prepend` method. """
        if isinstance(content, Selector):
            node.extend_before(0, content.data)
        elif isinstance(content, LC.Node):
            if (content.name in ['#document', '#document-fragment']
                    and content.temporary):
                node.extend_before(0, content)
            else:
                node.insert_before(0, content)
        elif hasattr(content, '__iter__'):
            node.extend_before(0, content)
        else:
            parser.parse(str(content))
            node.extend_before(0, parser.doc)

    def prepend(self, *arg, **keywords):
        """Insert content, specified by the parameter, to the
        beginning of each element in the set of matched elements.

        Should behave similarly as https://api.jquery.com/append/.
        Major difference is in the function. When passing a function
        it should take 2 parameters: node, index. Where node will be
        the current element to which the return value will be
        appended to. """
        info = {
            'lang': 'html',
            'style': 'default',
            'defaults': None,
        }
        for key in keywords:
            info[key] = keywords[key]
        parser = LC.Parser(info['lang'], info['style'], info['defaults'])
        if len(arg) == 1 and isinstance(arg[0], types.FunctionType):
            for num, node in enumerate(self.data):
                self._prepend(node, arg[0](node, num), parser)
        else:
            for content in arg:
                if isinstance(content, str):
                    parser.parse(content)
                    content = parser.doc
                elif isinstance(content, list):
                    for num in xrange(len(content)):
                        if isinstance(content[num], str):
                            parser.parse(content[num])
                            content[num] = parser.doc
                for i in xrange(len(self.data) - 1):
                    clone = clone_obj(content, parser)
                    self._prepend(self.data[i], clone, parser)
                if self.data:
                    self._prepend(self.data[-1], content, parser)

    @staticmethod
    def _after(node, content, parser):
        """Helper function to `after` method. """
        if isinstance(content, Selector):
            node.append_nodes_after(content.data)
        elif isinstance(content, LC.Node):
            if content.name in ['#document', '#document-fragment']:
                node.append_nodes_after(content)
            else:
                node.append_after(content)
        elif hasattr(content, '__iter__'):
            node.append_nodes_after(content)
        else:
            parser.parse(str(content))
            node.append_nodes_after(parser.doc)

    def after(self, *arg, **keywords):
        """Insert content, specified by the parameter, after each
        element in the set of matched elements.

        .. code::

            .after(content [,content])

        :param content:
            html string or Element or Array or jQuery string, Node,
            array of Node, or Selector object to insert after each
            element in the set of matched elements.

        :param content:
            html string or Element or Array or jQuery One or more
            additional DOM elements, arrays of elements, HTML
            strings, or jQuery objects to insert after each element
            in the set of matched elements.

        .. code::

            .after(function(node, index))

        :param function(node,index):
            A function that returns a string, DOM element(s), or
            Selector object to insert after each element in the set
            of matched elements. Receives the element in the set and
            its index position in the set as its arguments.

        .. code::

            .after(..., lang='html', style='default', 'defaults'=None)

        :param lang:
            The language in which strings will be parsed in.

        :param style:
            The style in which strings will be parsed in.

        :param defaults:
            A dictionary with string keywords and values specifying
            options for the particular style.
        """
        info = {
            'lang': 'html',
            'style': 'default',
            'defaults': None,
        }
        for key in keywords:
            info[key] = keywords[key]
        parser = LC.Parser(info['lang'], info['style'], info['defaults'])
        if len(arg) == 1 and isinstance(arg[0], types.FunctionType):
            for num, node in enumerate(self.data):
                self._after(node, arg[0](node, num), parser)
        else:
            for content in arg:
                if isinstance(content, str):
                    parser.parse(content)
                    content = parser.doc
                elif isinstance(content, list):
                    for num in xrange(len(content)):
                        if isinstance(content[num], str):
                            parser.parse(content[num])
                            content[num] = parser.doc
                for i in xrange(len(self.data) - 1):
                    clone = clone_obj(content, parser)
                    self._after(self.data[i], clone, parser)
                if self.data:
                    self._after(self.data[-1], content, parser)

    @staticmethod
    def _before(node, content, parser):
        """Helper function to `after` method. """
        if isinstance(content, Selector):
            node.prepend_nodes_before(content.data)
        elif isinstance(content, LC.Node):
            if content.name in ['#document', '#document-fragment']:
                node.prepend_nodes_before(content)
            else:
                node.prepend_before(content)
        elif hasattr(content, '__iter__'):
            node.prepend_nodes_before(content)
        else:
            parser.parse(str(content))
            node.prepend_nodes_before(parser.doc)

    def before(self, *arg, **keywords):
        """Insert content, specified by the parameter, before each
        element in the set of matched elements.

        .. code::

            .before(content [,content])

        :param content:
            html string or Element or Array or jQuery string, Node,
            array of Node, or Selector object to insert before each
            element in the set of matched elements.

        :param content:
            html string or Element or Array or jQuery One or more
            additional DOM elements, arrays of elements, HTML
            strings, or jQuery objects to insert before each element
            in the set of matched elements.

        .. code::

            .before(function(node, index))

        :param function(node,index):
            A function that returns a string, DOM element(s), or
            Selector object to insert before each element in the set
            of matched elements. Receives the element in the set and
            its index position in the set as its arguments.

        .. code::

            .before(..., lang='html', style='default', 'defaults'=None)

        :param lang:
            The language in which strings will be parsed in.

        :param style:
            The style in which strings will be parsed in.

        :param defaults:
            A dictionary with string keywords and values specifying
            options for the particular style.
        """
        info = {
            'lang': 'html',
            'style': 'default',
            'defaults': None,
        }
        for key in keywords:
            info[key] = keywords[key]
        parser = LC.Parser(info['lang'], info['style'], info['defaults'])
        if len(arg) == 1 and isinstance(arg[0], types.FunctionType):
            for num, node in enumerate(self.data):
                self._before(node, arg[0](node, num), parser)
        else:
            for content in arg:
                if isinstance(content, str):
                    parser.parse(content)
                    content = parser.doc
                elif isinstance(content, list):
                    for num in xrange(len(content)):
                        if isinstance(content[num], str):
                            parser.parse(content[num])
                            content[num] = parser.doc
                for i in xrange(len(self.data) - 1):
                    clone = clone_obj(content, parser)
                    self._before(self.data[i], clone, parser)
                if self.data:
                    self._before(self.data[-1], content, parser)

    def __iter__(self):
        """Iterate over the elements of the selector. """
        for node in self.data:
            yield node

    def __len__(self):
        """Return the number of elements.

        >>> x.__len__() == len(x)
        True

        """
        return len(self.data)



"""
Provides the |Parser| object which defines the basic mechanism for
parsing character sequences. This involves using objects derived from
the abstract class |NodeParser|.

.. |Parser| replace:: :class:`~.Parser`
.. |NodeParser| replace:: :class:`~.NodeParser`
.. |Node| replace:: :class:`~lexor.core.node.Node`
.. |Element| replace:: :class:`~lexor.core.elements.Element`
.. |PI| replace:: :class:`~lexor.core.elements.ProcessingInstruction`
.. |RawText| replace:: :class:`~lexor.core.elements.RawText`
.. |Void| replace:: :class:`~lexor.core.elements.Void`
.. |Document| replace:: :class:`~lexor.core.elements.Document`
.. |DocFrag| replace:: :class:`~lexor.core.elements.DocumentFragment`

"""
import re
import sys
from lexor.util import Position
from lexor.command import config
from lexor.command.lang import get_style_module, map_explanations
LC = sys.modules['lexor.core']


class NodeParser(object):
    """An object that has two methods: ``make_node`` and ``close``.
    Both method is required to be overloaded in derived objects."""

    def __init__(self, parser):
        """A |NodeParser| needs to be initialized with a |Parser|
        object. If this method is to be overloaded then make sure
        that it only accepts one parameter: ``parser``. This method
        is used by |Parser| and it calls it with itself as the
        parameter."""
        self.parser = parser

    def make_node(self):
        """This method is required to be overloaded by the derived
        node parser. It returns ``None`` if the node parser will not
        be able to create a node from the current information in the
        parser. Otherwise it creates a |Node| object and returns it.

        When returning a node you have the option of informing the
        parser if the node is complete or not. For instance, if your
        node parser creates an |Element| and it does not have any
        children to be parsed then return a ``list`` containing only
        the single node. This will tell the parser that the node has
        been closed and it will not call the :meth:`close` method of
        the node parser. If the |Node| does not have a child, say
        |PI|, |RawText|, or |Void| then there is no need to wrap the
        node in a ``list``.

        The |Node| object that this method returns also needs to have
        the property ``pos``. This is a ``list`` of two integers
        stating the line and column number where the node was
        encountered in the text that is being parsed. This property
        will be removed by the parser once the parser finishes all
        processing with the node.

        If this method is not overloaded as previously stated then a
        ``NotImplementedError`` exception will be raised.
        """
        msg = '%s did not implement `make_node`' % self.__class__
        raise NotImplementedError(msg)

    def close(self, _):
        """This method needs to be overloaded if the node parser
        returns a |Node| with the :meth:`make_node` method.

        This method will not get called if :meth:`make_node` returned
        a |Node| inside a ``list``. The close function takes as input
        the node object that :meth:`make_node` returned and it should
        decide if the node can be closed or not. If it is indeed time
        to close the node then return a ``list`` with the position
        where the node is being closed, otherwise return ``None``.

        If this method is not overloaded then a
        ``NotImplementedError`` exception will be raised.
        """
        msg = '%s did not implement `close`' % self.__class__
        raise NotImplementedError(msg)

    def msg(self, code, pos, arg=None, uri=None):
        """Send a message to the parser by providing one of the error
        codes defined in the style as well as the position where the
        error took place. The position has the form of a list
        containing the line and column number. Some error codes may
        provide arguments, this can be passed to `arg`. In case the
        error occurred somewhere not in the current document, perhaps
        in a string, then you may provide a new `uri` to denote the
        location."""
        self.parser.msg(self.__module__, code, pos, arg, uri)


# pylint: disable=too-many-instance-attributes
class Parser(object):
    """To see the languages that it is able to parse see the
    :mod:`lexor.command.lang` module. """

    def __init__(self, lang='xml', style='default', defaults=None):
        """Create a new parser by specifying the language and the
        style in which text will be parsed. """
        if defaults is None:
            defaults = dict()
        self._lang = lang
        self._style = style
        self._np = None
        self._next_check = None
        self._in_progress = None
        self._uri = None
        self._reload = True
        self.style_module = None
        self.text = None
        self.end = None
        self.pos = None
        self.caret = None
        self.doc = None
        self.log = None
        self.defaults = defaults
        self._node_parser = None
        self.current_node = None

    def _set_node_parser(self, val):
        """Helper function to create a node parser and store it
        in dictionary. """
        if isinstance(val, str):
            return self._node_parser[val]
        name = val.__name__
        self._node_parser[name] = val(self)
        return self._node_parser[name]

    def __getitem__(self, name):
        """Return the specified |NodeParser|. """
        return self._node_parser[name]

    def _set_node_parsers(self, lang, style, defaults=None):
        """Imports the correct module based on the language and style.
        """
        self.style_module = get_style_module('parser', lang, style)
        name = '%s-parser-%s' % (lang, style)
        config.set_style_cfg(self, name, defaults)
        self._next_check = dict()
        self._np = dict()
        self._node_parser = dict()
        if hasattr(self.style_module, 'REPOSITORY'):
            for val in self.style_module.REPOSITORY:
                self._set_node_parser(val)
        if hasattr(self.style_module, 'parser_setup'):
            self.style_module.parser_setup(self)
        str_key = list()
        for key, val in self.style_module.MAPPING.iteritems():
            self._next_check[key] = re.compile('.*?[%s]' % val[0])
            if isinstance(val, str):
                str_key.append((key, val))
            else:
                self._np[key] = [self._set_node_parser(p) for p in val[1]]
        for key, val in str_key:
            self._np[key] = self._np[val]

    def load_node_parsers(self):
        """Loads the |NodeParser| objects defined in a parser style.
        This function is called automatically when :meth:`parse` is
        called only if there was a change in the settings. """
        self._set_node_parsers(
            self._lang, self._style, self.defaults
        )
        self._reload = False

    def parse(self, text, uri=None):
        """Parses the given `text`. To see the results of this method
        see the :meth:`document` and :meth:`lexor_log` property. If
        no `uri` is given then :meth:`document` will return a
        |DocFrag| node. """
        if self._reload:
            self.load_node_parsers()
        self.text = text
        self.end = len(text)
        self.pos = [1, 1]
        self.caret = 0
        self.doc = LC.Document(self._lang)
        if uri:
            self._uri = uri
        else:
            self._uri = 'string@0x%x' % id(text)
        self.doc.uri_ = self._uri
        self.log = LC.Document("lexor", "log")
        self.log.modules = dict()
        self.log.explanation = dict()
        if hasattr(self.style_module, 'pre_process'):
            self.style_module.pre_process(self)
        self._parse()
        if hasattr(self.style_module, 'post_process'):
            self.style_module.post_process(self)
        map_explanations(self.log.modules, self.log.explanation)

    @property
    def cdata(self):
        """
        .. admonition:: Read-Only Property
            :class: note

            The character sequence data that was last processed by
            :meth:`parse`. This property is associated with the
            attribute ``text``."""
        return self.text

    @property
    def uri(self):
        """
        .. admonition:: Read-Only Property
            :class: note

            The Uniform Resource Identifier. This is the name that
            was given to the text that was last parsed."""
        return self._uri

    @property
    def position(self):
        """
        .. admonition:: Read-Only Property
            :class: note

            Position of caret in the text in terms of line and
            column. i.e. returns ``[line, column]``. This property is
            associated with the attribute ``pos``."""
        return self.pos

    @property
    def caret_position(self):
        """
        .. admonition:: Read-Only Property
            :class: note

            The index in the text the parser is processing. This
            property is associated with the attribute ``caret``."""
        return self.caret

    @property
    def lexor_log(self):
        """
        .. admonition:: Read-Only Property
            :class: note

            The index in the text the parser is processing. This
            property is associated with the attribute ``log``."""
        return self.log

    @property
    def document(self):
        """
        .. admonition:: Read-Only Property
            :class: note

            The parsed document. This is a |Document| or |DocFrag|
            created by the :meth:`parse` method. This property is
            associated with the attribute ``doc``."""
        return self.doc

    @property
    def language(self):
        """The language in which the |Parser| object will parse
        character sequences. """
        return self._lang

    @language.setter
    def language(self, value):
        """Setter function for style. """
        self._lang = value
        self._reload = True

    @property
    def parsing_style(self):
        """The style in which the |Parser| object will parse the
        character sequences. """
        return self._style

    @parsing_style.setter
    def parsing_style(self, value):
        """Setter function for style. """
        self._style = value
        self._reload = True

    def set(self, lang, style, defaults=None):
        """Set the language and style in one call. """
        if defaults is not None:
            self.defaults = defaults
        self._lang = lang
        self._style = style
        self._reload = True

    def copy_pos(self):
        """Return a copy of the current position. """
        return list(self.pos)

    def update(self, index):
        """Changes the position of the ``caret`` and updates ``pos``.
        This function assumes that you are moving forward. Do not
        update to an index which is less than the current position of
        the caret. """
        if index == self.caret:
            return
        nlines = self.text.count('\n', self.caret, index)
        self.pos[0] += nlines
        if nlines > 0:
            self.pos[1] = index - self.text.rfind('\n', self.caret, index)
        else:
            self.pos[1] += index - self.caret
        self.caret = index

    def update_log(self, log, after=True, delta=None):
        """Append the messages from a `log` document to the
        parsers log. This removes the children from `log`.
        """
        modules = log.modules
        explanation = log.explanation
        for mname in modules:
            if mname not in self.log.modules:
                self.log.modules[mname] = modules[mname]
            if mname not in self.log.explanation:
                self.log.explanation[mname] = explanation[mname]
        if delta:
            for error in log.iter_child_elements():
                error['uri'] = self.uri
                pos = error['position']
                pos[0] += delta[0]
                pos[1] += delta[1]
                for arg in error['arg']:
                    if isinstance(arg, Position):
                        arg.shift(delta)
        if after:
            self.log.extend_children(log)
        else:
            self.log.extend_before(0, log)

    def compute(self, index):
        """Return a position ``[line, column]`` in the text given an
        index. This does not modify anything in the parser. It only
        gives you the line and column where the caret would be given
        the index. The same applies as in update: do not use compute
        with an index less than the current position of the caret.
        """
        text = self.text
        nlines = text.count('\n', self.caret, index)
        tmpline = self.pos[0] + nlines
        if nlines > 0:
            tmpcolumn = index - text.rfind('\n', self.caret, index)
        else:
            tmpcolumn = self.pos[1] + index - self.caret
        return [tmpline, tmpcolumn]

    # pylint: disable=R0913
    def msg(self, mod_name, code, pos, arg=None, uri=None):
        """Provide the name of module issuing the message, the code
        number, the position of caret and optional arguments and uri.
        This information gets stored in the log.
        """
        if uri is None:
            uri = self._uri
        if arg is None:
            arg = ()
        node = LC.Void('msg')
        node['module'] = mod_name
        node['code'] = code
        node['position'] = list(pos)
        node['uri'] = uri
        node['arg'] = arg
        if mod_name not in self.log.modules:
            self.log.modules[mod_name] = sys.modules[mod_name]
        self.log.append_child(node)

    def _get_np(self, node):
        """Get a node parser based on the name of the node. """
        return self._np.get(node.name, self._np['__default__'])

    def _get_next_checker(self, node):
        """Get the checker based on the name of the node. """
        return self._next_check.get(
            node.name,
            self._next_check['__default__']
        )

    def _get_next_check(self, node):
        """Locate the index where a processor might return Node. If
        there is no index then return -1.
        """
        match = self._get_next_checker(node).search(
            self.text,
            self.caret
        )
        if match is None:
            return -1
        return match.end(0)-1

    def _process_node(self, crt, node, processor):
        """Appends the node to crt. """
        if isinstance(node, LC.Text):
            if len(crt) > 0 and isinstance(crt[-1], LC.Text):
                crt[-1].data += node.data
            elif node.data:
                crt.append_child(node)
        elif isinstance(node, list):  # Empty Element
            crt.append_child(node[0])
        else:
            crt.append_child(node)
            if isinstance(node.child, list):
                self._in_progress.append((node, processor))
                return node
        return None

    def _process_text(self, crt):
        """When there is no node then we just read the text. """
        index = self._get_next_check(crt)
        if index == -1:
            content = self.text[self.caret:self.end]
            if len(crt) > 0 and isinstance(crt[-1], LC.Text):
                crt[-1].data += content
            else:
                crt.append_child(content)
            self.update(self.end)
            return
        elif index - self.caret == 0:
            index += 1
        content = self.text[self.caret:index]
        self.update(index)
        if len(crt) > 0 and isinstance(crt[-1], LC.Text):
            crt[-1].data += content
        else:
            crt.append_child(content)

    def _close_node(self):
        """Checks and closes a node that is in self._in_progress. """
        num = len(self._in_progress)
        autoclose = None
        for node, processor in reversed(self._in_progress):
            num -= 1
            autoclose = processor.close(node)
            if autoclose is not None:
                break
        if autoclose is not None:
            # Must go backwards since the list `in_progress` is
            # changing.
            for i in xrange(len(self._in_progress)-1, num, -1):
                name = self._in_progress[i][0].name
                self.msg(
                    self.__module__, 'W100',
                    self._in_progress[i][0].pos,
                    (name, Position(autoclose))
                )
                del self._in_progress[i][0].pos
                del self._in_progress[i]
            del self._in_progress[num]
            if self._in_progress:
                return self._in_progress[-1][0]
            else:
                return self.doc
        return None

    def _parse(self):
        """Main parsing function. This function depends on the
        node parsers of the language. """
        self.current_node = crt = self.doc
        self._in_progress = []
        while self.caret < self.end:
            tmp = self._close_node()
            if tmp is not None:
                self.current_node = crt = tmp
                continue
            match = False
            processor = None
            for processor in self._get_np(crt):
                node = processor.make_node()
                if node is not None:
                    match = True
                    break
                elif self.caret == self.end:
                    break
            if match is False:
                self._process_text(crt)
            elif self._process_node(crt, node, processor) is node:
                self.current_node = crt = node
        for node, processor in self._in_progress:
            self.msg(self.__module__, 'E100', node.pos, [node.name])
            del node.pos


MSG = {
    'E100': 'closing string for `Node` of name "{0}" not found',
    'W100': 'auto-closing `Node` of name "{0}" at {1}',
}
MSG_EXPLANATION = [
    """
    - The parser did not find a closing string for the given node.

    - This is a general error which is language dependent. Make sure
      to provide the required closing string for the node.

    The following are examples for HTML, LaTeX and Lexor:

    Okay: <node></node>
    Okay: \\begin{node}\\end{node}
    Okay: %%{node}%%

    E100: <node>
    E100: \\begin{node}
    E100: %%{node}
""",
    """
    - The parser was forced to automatically close the current node
      in progress due to the encounter of the closing sequence of a
      parent node.

    - This is a general warning which is language dependent. To get
      rid of this warning provide the closing sequence for the node
      before the closing sequence of the parent node.

    The following is an example in HTML:

    Okay: <a><p>stuff</p><p>stuff</p></a>
    Okay: <a><p>stuff<p>stuff</p></a>

    W100: <a><p>stuff</p><p>stuff</a>
    W100: <a><p>stuff<p>stuff</a>
""",
]



import json
import httplib


def get(endpoint):
    """Make a get request to github. """
    connection = httplib.HTTPSConnection('api.github.com')
    connection.connect()
    connection.request('GET', endpoint, headers={
        'User-Agent': 'lexor'
    })
    return json.loads(connection.getresponse().read())



"""
Print the default values for each command.

"""
import textwrap
from lexor.command import import_mod
from lexor.command import config, LexorError


DESC = """
View default values for a subcommand.

"""


def add_parser(subp, fclass):
    """
    .. admonition:: Command Line Utility Function
        :class: warning

        Add a parser to the main subparser.
    """
    tmpp = subp.add_parser('defaults', help='print default values',
                           formatter_class=fclass,
                           description=textwrap.dedent(DESC))
    tmpp.add_argument('name', type=str, help='subcommand name')


def run():
    """
    .. admonition:: Command Line Utility Function
        :class: warning

        Run the command.
    """
    arg = config.CONFIG['arg']
    name = arg.name
    try:
        mod = import_mod('lexor.command.%s' % name)
    except ImportError:
        raise LexorError('invalid command: %r' % name)
    if hasattr(mod, 'DEFAULTS'):
        for key, val in mod.DEFAULTS.iteritems():
            print '%s = %r' % (key, val)
    else:
        print 'NO DEFAULTS'
